;-------------------------------------------------------------------------
;  Return of the Very Tiny Language
;  2006/04/26
;  Copyright (C) 2002-2006 Jun Mizutani <mizutani.jun@nifty.ne.jp>
;
;  file : vtlfb2.inc
;         frame buffer extention for rvtlc
;-------------------------------------------------------------------------

;               xor     ecx, ecx
;               mov     cl, 'a'
;               mov     ebp, [esi+ecx*4]
;               call    XXXXX

section .text

;---------------------------------------------------------------------------
; eax : Y
; ebx : width(byte)
; edi : addr
; carry : 32bits/pixel(set)
StartPoint:
                push    ebx
                imul    ebx                 ; Y * width
                mov     ebx, [ebp +  4]     ; X
                shl     ebx, 1              ; 16bit/pixel
                mov     edx, [ebp + 28]     ; depth
                cmp     dl, 32
                jne     .bit_16
                shl     ebx, 1              ; 32bit/pixel
                add     eax, ebx            ; Y * width + X
                add     edi, eax            ; edi := addr + eax
                pop     ebx
                stc
                ret

    .bit_16:    add     eax, ebx            ; Y * width + X
                add     edi, eax            ; edi := addr + eax
                pop     ebx
                clc
                ret

;---------------------------------------------------------------------------
; 饤    ebp=ؤΥݥ󥿤Υɥ쥹
; l[0] = addr   [ebp +  0]
; l[1] = x1     [ebp +  4]      ; l[2] = y1     [ebp +  8]
; l[3] = x2     [ebp + 12]      ; l[4] = y2     [ebp + 16]
; l[5] = color  [ebp + 20]
; l[6] = ScrX   [ebp + 24]
; l[7] = Depth  [ebp + 28]      ; 1ԥΥХȿ
; l[8] = incr1  [ebp + 32]
; l[9] = incr2  [ebp + 36]

LineDraw:
                pusha
                cld
                mov     edi, [ebp]          ; buffer address (mem or fb)
                mov     ebx, [ebp + 24]     ; ScrX

                mov     ecx, [ebp + 12]     ; ecx = delta X (X2 - X1)
                sub     ecx, [ebp +  4]     ; ecx = X2 - X1
                jz      .VertLine           ; if (delta X=0) Vertical
                jns     .L01                ; JUMP IF X2 > X1

                neg     ecx                 ; deltaX = - deltaX
                mov     eax, [ebp + 12]     ; swap X1  X2
                xchg    eax, [ebp +  4]
                mov     [ebp + 12], eax

                mov     eax, [ebp + 16]     ; swap Y1  Y2
                xchg    eax, [ebp +  8]     ; Y1
                mov     [ebp + 16], eax

    .L01:       mov     eax, [ebp + 16]     ; eax = Y2-Y1
                sub     eax, [ebp +  8]     ; Y1
                jnz     .SlopeLine

    .HolizLine: inc     ecx                 ; DELTA X + 1 : # OF POINTS
                mov     eax, [ebp +  8]     ; Y1
                call    StartPoint          ; edi=addr + X + Y * width
                mov     eax, [ebp + 20]     ; color
                jb      .h32
                rep     stosw               ; 16bit
                jmp     short .exit0        ; finished
    .h32:       rep     stosd               ; 32bit
                jmp     short .exit0        ; finished

    .VertLine:  mov     eax, [ebp +  8]     ; Y1
                mov     esi, [ebp + 16]     ; Y2
                mov     ecx, esi
                sub     ecx, eax            ; Y2 - Y1
                jge     .LVL1
                neg     ecx
                mov     eax, esi
    .LVL1:      inc     ecx                 ; DELTA Y + 1 : # OF POINTS
                call    StartPoint          ; edi=addr + X + Y * width
                mov     eax, [ebp + 20]     ; color
                jb      .LVL32
    .LVL16:
                mov     [edi], ax
                add     edi, ebx            ; ebx:width
                loop    .LVL16
                jmp     short .exit0        ; finished
    .LVL32:
                mov     [edi], eax
                add     edi, ebx            ; ebx:width
                loop    .LVL32
    .exit0:     popa
                ret                         ; finished

    ;-------------------------------------------------
    ;      ENTRY : eax = DY   ebx = width (bytes/line)
    ;              ecx = dx
    .SlopeLine:
                jns     .L03                ; JUMP IF (Y2 > Y1)
                neg     eax                 ; DELTA Y
                neg     ebx                 ; BYTES/LINE

    .L03:       push    eax
                push    ecx
                cmp     eax, ecx            ; DELTA Y - DELTA X
                jle     .L04                ; JUMP IF DY <= dx ( SLOPE <= 1)
                xchg    eax, ecx

    .L04:       shl     eax, 1              ; eax := 2 * DY
                mov     [ebp + 32], eax     ; incr1 := 2 * DY
                sub     eax, ecx
                mov     esi, eax            ; esi := D = 2 * DY - dx
                sub     eax, ecx

                mov     [ebp + 36], eax     ; incr2 := D = 2 * (DY - dx)

                mov     eax, [ebp +  8]     ; Y1
                push    ebx
                or      ebx, ebx
                jns     .L05
                neg     ebx
    .L05:
                call    StartPoint          ; edi=addr + X + Y * width
                pop     ebx
                pop     ecx
                pop     eax
                jb      .Slope32
                cmp     eax, ecx            ; DELTA Y - DELTA X
                jg      .HiSlopeLine        ; JUMP IF DY > dx ( SLOPE > 1)

    .LoSlopeLine:
                inc     ecx
                mov     eax, [ebp + 20]     ; color
    .L11:       stosw
                or      esi, esi
                jns     .L12
                add     esi, [ebp + 32]     ; incr1
                loop    .L11
                jmp     short .exit

    .L12:       add     esi, [ebp + 36]     ; incr2
                add     edi, ebx            ; ebx=(+/-)width
                loop    .L11
                jmp     short .exit

    .HiSlopeLine:
                xchg    eax, ecx
                inc     ecx
                mov     eax, [ebp + 20]     ; color
    .L21:       stosw
                add     edi, ebx            ; ebx=(+/-)width
                or      esi, esi
                jns     .L22
                add     esi, [ebp + 32]     ; incr1
                dec     edi
                dec     edi
                loop    .L21
                jmp     short .exit

    .L22:       add     esi, [ebp + 36]     ; incr2
                loop    .L21

    .exit:      popa
                ret

    .Slope32:   cmp     eax, ecx            ; DELTA Y - DELTA X
                jg      .HiSlopeLine32      ; JUMP IF DY > dx ( SLOPE > 1)

    .LoSlopeLine32:
                inc     ecx
                mov     eax, [ebp + 20]     ; color
    .L31:       stosd
                or      esi, esi
                jns     .L32
                add     esi, [ebp + 32]     ; incr1
                loop    .L31
                jmp     short .exit2

    .L32:       add     esi, [ebp + 36]     ; incr2
                add     edi, ebx            ; ebx=(+/-)width
                loop    .L31
                jmp     short .exit2

    .HiSlopeLine32:
                xchg    eax, ecx
                inc     ecx
                mov     eax, [ebp + 20]     ; color
    .L41:       stosd
                add     edi, ebx            ; ebx=(+/-)width
                or      esi, esi
                jns     .L42
                add     esi, [ebp + 32]     ; incr1
                sub     edi, 4
                loop    .L41
                jmp     short .exit2

    .L42:       add     esi, [ebp + 36]     ; incr2
                loop    .L41

    .exit2:     popa
                ret

;---------------------------------------------------------------------------
;  16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;   d[0] = addr   [ebp +  0] ΰƬɥ쥹
;   d[1] = x      [ebp +  4] Xɸ
;   d[2] = y      [ebp +  8] Yɸ
;   d[3] = Color  [ebp + 12] 
;   d[4] = ScrX   [ebp + 16] žXΥХȿ
;   d[5] = Depth  [ebp + 20] 1ԥΥХȿ
Dot:
                pusha
                mov     ecx, [ebp + 16]     ; ScrX
                mov     edi, [ebp]          ; buffer address (mem or fb)
                mov     eax, [ebp +  8]     ; Y
                imul    ecx                 ; Y * ScrX
                mov     edx, [ebp +  4]     ; X
                shl     edx, 1              ; X * 2
                add     eax, edx            ; X * 2 + Y * ScrX
                add     edi, eax            ; edi := addr + eax
                mov     ebx, [ebp + 12]     ; color
                mov     eax, [ebp + 20]     ; Depth
                cmp     al, 32
                je      .depth32
                mov     [edi], bx           ; 16bit/pixel
                popa
                ret
    .depth32:   mov     [edi], ebx          ; 32bit/pixel
                popa
                ret


;---------------------------------------------------------------------------
; Ϣ̽     ebp=ؤΥݥ󥿤Υɥ쥹
; return
;  eax = depth
;  ebx = ScrX
;  ecx = ScrX
;  edx = PatH
;  edi = ž襢ɥ쥹
;  esi = ѥγǼɥ쥹Ƭ
;  ebp = Ƭ
PatternSize:
                cld
                mov     ecx, [ebp + 24]     ; ScrX
                mov     edi, [ebp]          ; buffer address (mem or fb)
                mov     esi, [ebp + 20]     ; pattern
                mov     eax, [ebp +  8]     ; Y
                imul    ecx                 ; Y * ScrX
                mov     edx, [ebp +  4]     ; X
                shl     edx, 1              ; X * 2
                add     eax, edx            ; X * 2 + Y * ScrX
                add     edi, eax            ; edi := addr + eax
                mov     ebx, ecx            ; ebx := Screen Width
                mov     edx, [ebp + 16]     ; PatH
                mov     al,  [ebp + 28]     ; Depth
                ret

;---------------------------------------------------------------------------
; ѥž 16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;   p[0] = addr   [ebp +  0] ž襢ɥ쥹
;   p[1] = x      [ebp +  4] žXɸ
;   p[2] = y      [ebp +  8] žYɸ
;   p[3] = PatW   [ebp + 12] ѥ
;   p[4] = PatH   [ebp + 16] ѥι⤵
;   p[5] = mem    [ebp + 20] ѥγǼɥ쥹
;   p[6] = ScrX   [ebp + 24] žXΥХȿ
;   p[7] = Depth  [ebp + 28] 1ԥΥХȿ

PatternTransfer:
                pusha
                call    PatternSize
                cmp     al, 32
                je      .pat32
    .loop16:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                rep     movsw
                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop16
                jmp     short .exit
    .pat32:     add     edi, edx            ; edi=addr + Y*ScrX + X*4
    .loop32:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                rep     movsd
                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop32
    .exit:      popa
                ret

;---------------------------------------------------------------------------
; ѥž2 16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;   t[0] = addr   [ebp +  0] ž襢ɥ쥹
;   t[1] = x      [ebp +  4] žXɸ
;   t[2] = y      [ebp +  8] žYɸ
;   t[3] = PatW   [ebp + 12] ѥ
;   t[4] = PatH   [ebp + 16] ѥι⤵
;   t[5] = mem    [ebp + 20] ѥγǼɥ쥹Ƭ
;   t[6] = ScrX   [ebp + 24] žXΥХȿ
;   t[7] = Depth  [ebp + 28] 1ԥΥХȿ
;   t[8] = x2     [ebp + 32] žXɸ
;   t[9] = y2     [ebp + 36] žYɸ
;   t[10]= ScrX2  [ebp + 40] žXΥХȿ

PatternTransfer2:
                pusha
                call    PatternSize
                push    eax                 ; depth
                push    edx                 ; PatH
                mov     ecx, [ebp + 40]     ; ScrX2
                mov     eax, [ebp + 36]     ; Y2
                imul    ecx                 ; Y2 * ScrX2
                mov     edx, [ebp + 32]     ; X2
                shl     edx, 1              ; X2 * 2
                add     eax, edx            ; X2 * 2 + Y2 * ScrX2
                add     esi, eax            ; esi := mem + eax
                pop     edx                 ; PatH
                pop     eax                 ; depth
                cmp     al, 32
                je      .pat32
                mov     eax, [ebp + 40]     ; ScrX2
    .loop16:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                push    esi
                rep     movsw
                pop     esi
                pop     edi
                add     edi, ebx            ; y++
                add     esi, eax            ; y2++
                dec     edx
                jne     .loop16
                jmp     short .exit
    .pat32:     add     edi, edx            ; edi=addr + Y*ScrX + X*4
                mov     eax, [ebp + 40]     ; ScrX2
    .loop32:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                push    esi
                rep     movsd
                pop     esi
                pop     edi
                add     edi, ebx            ; y++
                add     esi, eax            ; y2++
                dec     edx
                jne     .loop32
    .exit:      popa
                ret

;---------------------------------------------------------------------------
; ޥդѥž 16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;   q[0] = addr   [ebp +  0] ž襢ɥ쥹
;   q[1] = x      [ebp +  4] žXɸ
;   q[2] = y      [ebp +  8] žYɸ
;   q[3] = PatW   [ebp + 12] ѥ
;   q[4] = PatH   [ebp + 16] ѥι⤵
;   q[5] = mem    [ebp + 20] ѥγǼɥ쥹
;   q[6] = ScrX   [ebp + 24] XΥХȿ
;   q[7] = Depth  [ebp + 28] 1ԥΥХȿ
;   q[8] = Mask   [ebp + 32] ޥ
MPatternTransfer:
                pusha
                call    PatternSize
                cmp     al, 32
                je      .pat32
    .loop16:    mov     ecx, [ebp + 12]     ; PatW
                push    edi

    .l16b       mov     ax, [esi]
                cmp     ax, [ebp + 32]      ; ޥʤޤʤ
                je      .l16n
                mov     [edi], ax
    .l16n       add     esi, 2
                add     edi, 2
                loop    .l16b

                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop16
                jmp     short .exit

    .pat32:     add     edi, edx            ; edi=addr + Y*ScrX + X*4
    .loop32:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
    .l32b       mov     eax, [esi]
                cmp     eax, [ebp + 32]     ; ޥʤޤʤ
                je      .l32n
                mov     [edi], eax
    .l32n       add     esi, 4
                add     edi, 4
                loop    .l32b
                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop32
    .exit:      popa
                ret

;---------------------------------------------------------------------------
; ѥե 16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;   r[0] = addr   [ebp +  0] ž襢ɥ쥹
;   r[1] = x      [ebp +  4] žXɸ
;   r[2] = y      [ebp +  8] žYɸ
;   r[3] = PatW   [ebp + 12] ѥ
;   r[4] = PatH   [ebp + 16] ѥι⤵
;   r[5] = Color  [ebp + 20] ѥο
;   r[6] = ScrX   [ebp + 24] XΥХȿ
;   r[7] = Depth  [ebp + 28] 1ԥΥХȿ

PatternFill:
                pusha
                call    PatternSize
                cmp     al, 32
                je      .pat32
    .pat16:     mov     eax, esi
    .loop16:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                rep     stosw
                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop16
                jmp     short .exit
    .pat32:     mov     eax, esi
                add     edi, edx            ; edi=addr + Y*ScrX + X*4
    .loop32:    mov     ecx, [ebp + 12]     ; PatW
                push    edi
                rep     stosd
                pop     edi
                add     edi, ebx            ; y++
                dec     edx
                jne     .loop32
    .exit:      popa
                ret

;---------------------------------------------------------------------------
; ե 8&16&32bit     ebp=ؤΥݥ󥿤Υɥ쥹
;  m[0] = addr   [ebp +  0] եƬɥ쥹
;  m[1] = offset [ebp +  4] եå
;  m[2] = length [ebp +  8] Ĺ(ԥñ)
;  m[3] = color  [ebp + 12] 
;  m[4] = Depth  [ebp + 16] bits/pixel

FrameBufferFill:
                pusha
                cld
                mov     edx, ebp
                mov     edi, [edx]
                add     edi, [edx + 4]      ; offset
                mov     ecx, [edx + 8]      ; length (pixel)
                mov     eax, [edx +12]      ; color
                mov     edx, [edx +16]      ; bits/pixel
                shr     edx, 4
                jne     .word
                rep     stosb               ; byte
                jmp     short .exit
    .word:      shr     edx, 1
                jne     .dword
                rep     stosw               ; word
                jmp     short .exit
    .dword:     rep     stosd               ; dword
    .exit:      popa
                ret

;---------------------------------------------------------------------------
; ꥳԡ       ebp=ؤΥݥ󥿤Υɥ쥹
;  c[0] = source [ebp +  0] žƬɥ쥹
;  c[1] = dest   [ebp +  4] žƬɥ쥹
;  c[2] = length [ebp +  8] žХȿ
MemCopy:
                pusha
                cld
                mov     edx, ebp
                mov     esi, [edx]          ; žɥ쥹
                mov     edi, [edx + 4]      ; ž襢ɥ쥹
                mov     ecx, [edx + 8]      ; žХȿ
                mov     eax, edi
                sub     eax, esi            ; eax = ž - ž
                jle     .normal             ; ž < ž
                std                         ; ս
                dec     ecx                 ; Բ᤮˽
                add     esi, ecx            ; Ǹ
                add     edi, ecx
                inc     ecx                 ; ž̤᤹
    .normal:
                rep     movsb               ; Хñ
                cld
                popa
                ret
